/*  EQEMu:  Everquest Server Emulator
    Copyright (C) 2001-2007  EQEMu Development Team (http://eqemu.org)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY except by those people which sell it, which
	are required to give you total support for your newly bought product;
	without even the implied warranty of MERCHANTABILITY or FITNESS FOR
	A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "../common/debug.h"
#include "../common/moremath.h"
#include "../common/packet_dump_file.h"
#include "zone.h"
#include "npc.h"
#include "NpcAI.h"
#include "map.h"
#include "entity.h"
#include "masterentity.h"
#include "spdat.h"
#include "ldonchests.h"
#include "../common/bodytypes.h"
#include "../common/ldonchesttypes.h"
#include "spawngroup.h"
#include "../common/MiscFunctions.h"
#include "../common/rulesys.h"
#include "StringIDs.h"

void	LDoNChest_Handle_Open(Client* client, NPC* chest)
{
	NPC * target = chest;
	if(target && client)
	{
		if(target->IsNPC() && target->IsLDoNChest())
		{

			// have to be close to the chest
			if(client->DistNoRootNoZ(*target) > 80)
			{
				client->Message_StringID(13, TARGET_TOO_FAR);
				return;
			}

			if(target->LDoNChest_IsTrapped())
			{
				client->Message_StringID(13, LDT_ACCIDENT_SETOFF2);
				target->LDoNChest_TriggerTrap(client);
				return; // must not fall through when triggering a trap or the effect won't land
			}

			if(target->LDoNChest_IsLocked())
			{
				client->Message_StringID(MT_Skills, LDT_STILL_LOCKED, target->GetCleanName());
			}
			else
				target->LDoNChest_Open(client);
		}
	}
}

void	LDoNChest_Handle_Appraise(Client* client, NPC* chest, int clientskill, int failmsgid)
{
	NPC * target = chest;
	if(target && client)
	{
		if(target->IsNPC() && target->IsLDoNChest())
		{
			int check=LDoNChest_SkillCheck(target, clientskill);

			if(target->IsLDoNChest_TrappedType() && target->LDoNChest_IsTrapped()==false) // the trap was tiggered, we're certain it is no longer a trap now
				client->Message_StringID(MT_Skills, LDT_CERTAIN_NOT_TRAP, target->GetCleanName());
			else
			{
				if(target->IsLDoNChest_TrappedType() && target->LDoNChest_TrapSensed()==true) // we detected it previously
					client->Message_StringID(MT_Skills, LDT_CERTAIN_TRAP, target->GetCleanName());
				else
				{
					switch(check)
					{
					case -1: // critical failure; fortunately this won't trigger the trap
					case 0: // failed
						client->Message_StringID(MT_Skills, failmsgid, target->GetCleanName());
						break;
					case 1: // success
						if(target->LDoNChest_IsTrapped()==true)
						{
							target->LDoNChest_Sensed(); // player sensed trap
							client->Message_StringID(MT_Skills, LDT_CERTAIN_TRAP, target->GetCleanName());
						}
						else
							client->Message_StringID(MT_Skills, failmsgid, target->GetCleanName());
						break;
					}
				}
			}
		}
	}
}

void	LDoNChest_Handle_Disarm(Client* client, NPC* chest, int clientskill)
{
	NPC * target = chest;
	if(target && client)
	{
		if(target->IsNPC() && target->IsLDoNChest())
		{
			int check=0;

			if(target->LDoNChest_TrapSensed()==true)
				check=LDoNChest_SkillCheck(target, clientskill);
			else
				check=LDoNChest_SkillCheck(target, clientskill/2); // didn't sense trap before disarming - that's risky

			if(target->IsLDoNChest_TrappedType() && target->LDoNChest_IsTrapped()==false) // the trap was tiggered, we're certain it is no longer trapped now
				client->Message_StringID(MT_Skills, LDT_WAS_NOT_TRAPPED, target->GetCleanName());
			else
			{
				switch(check)
				{
				case 0: // failed
					client->Message_StringID(MT_Skills, LDT_HAVE_NOT_DISARMED, target->GetCleanName());
					break;
				case 1: // success
					if(target->LDoNChest_IsTrapped()==true)
					{
						target->LDoNChest_Disarm(); // player disarmed trap
						client->Message_StringID(MT_Skills, LDT_HAVE_DISARMED, target->GetCleanName());
					}
					else
						client->Message_StringID(MT_Skills, LDT_WAS_NOT_TRAPPED, target->GetCleanName());
					break;
				case -1: // critical failure
					if(target->LDoNChest_IsTrapped()==true)
					{
						client->Message_StringID(13, LDT_ACCIDENT_SETOFF);
						target->LDoNChest_TriggerTrap(client);
					}
					else
						client->Message_StringID(MT_Skills, LDT_HAVE_NOT_DISARMED, target->GetCleanName());
					break;
				}
			}
		}
	}
}

void	LDoNChest_Handle_Picklock(Client* client, NPC* chest, int clientskill)
{
	NPC * target = chest;
	if(target && client)
	{
		if(target->IsNPC() && target->IsLDoNChest())
		{
			int check=LDoNChest_SkillCheck(target, clientskill);

			// if it's still trapped - not good...
			if(target->LDoNChest_IsTrapped()==true)
			{
				client->Message_StringID(13, LDT_ACCIDENT_SETOFF2);
				target->LDoNChest_TriggerTrap(client);
				return;
			}

			if(target->IsLDoNChest_LockedType() && target->LDoNChest_IsLocked()==false) // it's a locked type, but it was unlocked already
				client->Message_StringID(MT_Skills, LDT_WAS_NOT_LOCKED, target->GetCleanName());
			else
			{
				switch(check)
				{
				case -1: // critical failure
				case 0: // failed
					client->Message_StringID(MT_Skills, LDT_PICKLOCK_FAILURE, target->GetCleanName());
					break;
				case 1: // success
					if(target->LDoNChest_IsLocked()==true)
					{
						target->LDoNChest_Unlock(); // player opened the lock
						client->Message_StringID(MT_Skills, LDT_PICKLOCK_SUCCESS, target->GetCleanName());
					}
					else
						client->Message_StringID(MT_Skills, LDT_WAS_NOT_LOCKED, target->GetCleanName());
					break;
				}
			}
		}
	}

}
int		LDoNChest_SkillCheck(NPC* chest, int skill)
{
	int		ret=0;
	int		chest_difficulty=chest->GetLevel()*5;
	float	base_difficulty=15.0f; // set this higher to make all disarm/picklock attempts more difficult

	// minimum a level 1 chest
	if(chest_difficulty==0)
		chest_difficulty=5;

	// cout << "Player skill: " << skill << endl;
	// cout << "Chest difficulty: " << chest_difficulty << endl;

	float	chance=((100.0f-base_difficulty)*((float)skill/(float)chest_difficulty)); // base difficulty; if skill=difficulty, a XX% chance of failure remains

	if(chance>100.0f-base_difficulty)
	{
		chance=100.0f-base_difficulty; // cap
		// cout << "Capping chance at: " << chance << endl;
	}

	// cout << "Chance to open: " << chance << endl;

	float	d100=(float)MakeRandomFloat(0, 85)+base_difficulty;

	// cout << "D100 roll: " << d100 << endl;

	if(d100<=chance)
		ret=1;
	else
	{
		if(d100 > (chance+10.0f))
			ret=-1; // critical failure
	}

	// cout << "Check=" << ret << endl;
	return(ret);
}
